#!/usr/bin/env python3
"""
File encrypt/decrypt (AES-256-CBC with HMAC-SHA256)
"""

import os
import sys
import logging
import string
from io import BytesIO
from pathlib import Path

from envex.env_crypto import encrypt_data, decrypt_data, DecryptError

ENCRYPTED_EXT = ".enc"

logging.basicConfig(
    format="%(asctime)s %(levelname).3s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
    force=True,
    handlers=[logging.StreamHandler(sys.stdout)],
)
logger = logging.getLogger(__file__)


def check_password_simple(password: str) -> bool:
    """
    Check if a password is too simple based on length,
    character variety, and basic complexity rules.
    Returns True if the password is too simple.
    """
    # Minimum length check
    if len(password) < 8:
        return True

    # Check for character variety
    has_upper = any(char.isupper() for char in password)
    has_lower = any(char.islower() for char in password)
    has_digit = any(char.isdigit() for char in password)
    all_digit = all(char.isdigit() for char in password)
    has_special = any(char in string.punctuation for char in password)

    if not has_upper or not has_lower or not has_digit or not has_special or all_digit:
        return True

    # Check for repeated/sequential characters
    if len(set(password)) <= 3:  # Mostly repeated characters
        return True

    repeated_patterns = ["12345", "abcde", "password", "qwerty", "asdf"]
    return any((pattern in password.lower() for pattern in repeated_patterns))


def main():
    import io
    import argparse

    class CustomParser(argparse.ArgumentParser):
        def error(self, message):
            if message:
                logger.error(f"{self.prog}: error: {message}")
            self.print_usage(sys.stderr)
            self.exit(2)

        def print_usage(self, file = None):
            text = io.StringIO()
            self._print_message(self.format_usage(), text)
            self.print_text(text)

        def print_text(self, text):
            text.seek(0)
            for line in text.readlines():
                logger.info(line.strip())

        def warning(self, message):
            if message:
                logger.warning(f"{self.prog}: warning: {message}")

        def print_help(self, file=None):
            text = io.StringIO()
            super().print_help(file=text)
            self.print_text(text)

    class CustomHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
        def __init__(self, *args, **kwargs):
            kwargs["max_help_position"] = 45
            kwargs["width"] = 1000
            super().__init__(*args, **kwargs)


    parser = CustomParser(description=__doc__, formatter_class=CustomHelpFormatter)
    password_opts = parser.add_mutually_exclusive_group()
    password_opts.add_argument(
        "-P", "--password", action="store", help="Use given password"
    )
    password_opts.add_argument(
        "-E",
        "--environ",
        action="store",
        help="Read password from provided environment variable",
    )
    password_opts.add_argument(
        "-F", "--file", action="store", help="Read password from a given file"
    )

    encrypt_opts = parser.add_mutually_exclusive_group()
    encrypt_opts.add_argument(
        "-e", "--encrypt", action="store_true", default=False, help="Use given password"
    )
    encrypt_opts.add_argument(
        "-d",
        "--decrypt",
        action="store_true",
        default=False,
        help="Read password from provided environment variable",
    )

    parser.add_argument(
        "-r",
        "--rm",
        action="store_true",
        default=False,
        help="Remove input file after successful conversion",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Increase output verbosity",
    )
    parser.add_argument("input", action="store", help="File to encrypt or decrypt")
    parser.add_argument(
        "output", nargs="?", default=None, action="store", help="Output file (optional)"
    )

    args = parser.parse_args()

    if not (_password := args.password):
        if args.environ:
            _password = os.environ.get(args.environ)
            if not _password:
                logger.error(f"{parser.prog}: {args.environ} is not set or is empty")
                exit(2)
        elif args.file:
            _password = Path(args.file).read_text()
        elif sys.stdin.isatty():
            import getpass

            _password = getpass.getpass(f"{parser.prog}: Enter passphrase: ")
        else:
            logger.error(f"{parser.prog}: password or password source is not provided")
            exit(2)

    if check_password_simple(_password):
        logger.warning("""WARNING: Password appears to be too short, simple or can easily be guessed.
  Recommended:
    - at least 1 each of uppercase, lowercase, digit, punctuation
    - at least 8 characters in length
    - does not contain common sequences used in passwords""")

    if not args.encrypt and not args.decrypt:
        logger.error(
            "{parser.prog}: operation to perform (--encrypt or --decrypt) must be specified"
        )
        exit(3)

    encrypt = args.encrypt

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    _input = Path(args.input)
    if not _input.exists():
        logger.error(f"{parser.prog}: {_input} does not exist")
        exit(1)

    if not args.output:
        if encrypt:
            args.output = f"{args.input}{ENCRYPTED_EXT}"
        elif args.input.endswith(ENCRYPTED_EXT):
            args.output = args.input[: -len(ENCRYPTED_EXT)]
        else:
            logger.error(
                f"{parser.prog}: cannot automatically determine output file name"
            )
            exit(1)

    _output = Path(args.output)
    if _output.exists() and _input.samefile(_output):
        logger.error(f"{parser.prog}: Input and Output files cannot be the same")
        exit(1)

    logger.debug(
        f"{parser.prog}: {'encrypt' if encrypt else 'decrypt'} {_input} -> {_output}"
    )

    func = encrypt_data if encrypt else decrypt_data
    try:
        _output.write_bytes(func(BytesIO(_input.read_bytes()), _password).getvalue())
    except DecryptError as e:
        logger.error(f"{parser.prog}: {e.args}")
        exit(4)

    if args.rm:
        logger.debug(f"{parser.prog}: removing {_input}")
        _input.unlink(missing_ok=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Import variables from a .env file to hashicorp vault.
"""

import logging
import os

import envex
from envex.env_hvac import SecretsManager

logging.captureWarnings(True)


def expand(p: str):
    return os.path.expandvars(os.path.expanduser(p))


def handler(
    files: list[str],
    url: str = None,
    token: str = None,
    cert: tuple = None,
    verify: bool | str = True,
    unseal: str = None,
    namespace: str = None,
    environ: str = None,
):
    sm = SecretsManager(url=url, token=token, cert=cert, verify=verify)
    if unseal:
        sm.unseal(keys=unseal.split(","), root_token=token)

    if sm.client is None:
        # noinspection PyArgumentList
        logging.fatal(
            "Can't connect or authenticate with Vault", exc_info=False, exitcode=1
        )

    if sm.client.seal_status["sealed"]:
        # noinspection PyArgumentList
        logging.fatal("Vault is currently sealed", exc_info=False, exitcode=4)

    try:
        path = sm.join(namespace, environ)
        for filename in files:
            filename = expand(filename)
            try:
                env = envex.Env(
                    readenv=True,
                    environ={},
                    env_file=filename,
                    update=False,
                    errors=False,
                    # pass these on in case we need them for completion
                    url=url,
                    token=token,
                    cert=cert,
                    verify=verify,
                    base_path=path,
                )
                count = 0
                secrets = {}
                for k, v in env.items():
                    if k not in ("CWD", "PWD"):
                        key = sm.join(path, k)
                        if v is not None:
                            secrets[key] = v
                            count += 1
                logging.info(
                    f"Added or updated {count} items from {filename} to '{path}'"
                )
            except IOError as e:
                logging.error(f"{filename}: {e.__class__.__name__}", exc_info=True)
    finally:
        # reseal the vault
        if unseal:
            sm.seal()


def main():
    import argparse

    from envex.scripts.lib.decr_action import Decrement
    from envex.scripts.lib.log import config as log_config
    from envex.scripts.lib.log import log_get_level, log_set_level

    log_config()

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-n",
        "--namespace",
        type=str,
        default="myapp",
        help="Namespace or application name",
    )
    parser.add_argument(
        "-e",
        "--environ",
        type=str,
        default=None,
        help="The code environment used to create or update variables",
    )
    parser.add_argument(
        "-a",
        "--address",
        type=str,
        default=None,
        help="The address/url of the hashicorp vault server",
    )
    parser.add_argument(
        "-t",
        "--token",
        type=str,
        default=None,
        help="The token used to authenticate to hashicorp vault",
    )
    parser.add_argument(
        "-u",
        "--unseal",
        default="",
        help="Unseal/reseal the vault with the provided comma-separated list of key",
    )
    parser.add_argument(
        "-c",
        "--cert",
        type=str,
        default=None,
        help="Client cert (if any)",
    )
    parser.add_argument(
        "-k",
        "--key",
        type=str,
        default=None,
        help="Client cert key (if any)",
    )
    parser.add_argument(
        "-N",
        "--noverify",
        dest="verify",
        action="store_false",
        default=True,
        help="Disable server certificate verification",
    )
    parser.add_argument(
        "-C",
        "--cacert",
        dest="verify",
        default=True,
        help="Path to a custom CA certificate (do not use with -N)",
    )
    parser.add_argument(
        "files",
        nargs="+",
        help="Filename(s) from which to read variables.",
    )
    default_level = log_get_level()
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=default_level,
        dest="verbose",
        help="Verbose output (specify multiple times for more verbosity)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action=Decrement,
        default=default_level,
        dest="verbose",
        help="Verbose output (specify multiple times for more verbosity)",
    )
    args = parser.parse_args()
    certs = (args.cert, args.key) if args.cert and args.key else None

    log_set_level(args.verbose)

    handler(
        args.files,
        url=args.address,
        token=args.token,
        cert=certs,
        verify=args.verify,
        unseal=args.unseal,
        namespace=args.namespace,
        environ=args.environ,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import os
import sys
import contextlib
import re
from io import TextIOBase, BytesIO, BufferedReader
from pathlib import Path
from typing import (
    Dict,
    List,
    MutableMapping,
    Union,
    Optional,
    Any,
    Generator,
)

from .env_crypto import decrypt_data, DecryptError

__all__ = (
    "load_env",
    "load_stream",
    "load_dotenv",  # alias
    "update_env",
    "unquote",
)


DEFAULT_ENVKEY = "DOTENV"
DEFAULT_DOTENV = ".env"
ENCRYPTED_EXT = ".enc"
DEFAULT_ENCODING = "utf-8"

# Precompiled regular expression patterns
_MODIFIER_PATTERN = re.compile(r":([-+])")
_VAR_BRACES_PATTERN = re.compile(r"\${([^{}]+)}")
_VAR_NO_BRACES_PATTERN = re.compile(r"\$([a-zA-Z_][a-zA-Z0-9_]*)")


def unquote(line, quotes="\"'"):
    if line and line[0] in quotes and line[-1] == line[0]:
        line = line[1:-1]
    return line


def update_env(env: MutableMapping[str, str], mapping: Dict):
    for k, v in mapping.items():
        env[str(k)] = str(v)


def _env_default(
    environ: MutableMapping[str, str], key: str, val: str, overwrite: bool = False
):
    if key and val and (overwrite or key not in environ):
        environ[key] = val


def _env_export(
    environ: MutableMapping[str, str], key: str, val: str, overwrite: bool = False
):
    if key and val and (overwrite or key not in environ):
        environ[key] = val
        os.environ[key] = val


def _env_files(
    env_file: str, search_path: List[Path], parents: bool, decrypt: bool, errors: bool
) -> Generator[str | None, Any, None]:
    """expand env_file with the full search path, optionally parents as well"""

    def resolve_file(base_path: Path, name: str, _decrypt: bool) -> Optional[str]:
        """Returns the path to the env file, prioritising the encrypted version if enabled"""
        if _decrypt:
            encrypted_path = os.path.join(base_path, name + ENCRYPTED_EXT)
            if os.access(encrypted_path, os.R_OK):
                return encrypted_path

        standard_path = os.path.join(base_path, name)
        return standard_path if os.access(standard_path, os.R_OK) else None

    searched = []
    for path in search_path:
        path = path.resolve()
        if not path.is_dir():
            path = path.parent
        searched.append(path)

        for sub_path in [path] + list(path.parents):
            if env_path := resolve_file(sub_path, env_file, decrypt):
                yield env_path
                break
            elif not parents:
                break

    if errors:
        raise FileNotFoundError(f"{env_file} in {[s.as_posix() for s in searched]}")
    else:
        yield env_file


@contextlib.contextmanager
def open_env(path: Union[str, Path]) -> Generator[BufferedReader, Any, None]:
    """same as open, allow monkeypatch"""
    fp = open(path, "rb")
    try:
        yield fp
    finally:
        fp.close()


ENV_COMMANDS = {
    "export": _env_export,
}


def _process_line(_lineno: int, string: str, errors: bool, _env_path: Path | None):
    """process a single line"""
    _func, _key, _val = _env_default, None, None
    parts = string.split("=", 1)
    if len(parts) == 2:
        _key, _val = parts
    elif len(parts) == 1:
        _key = parts[0]
    if _key:
        words = _key.split(maxsplit=1)
        if len(words) > 1:
            command, _key = words
            try:
                _func = ENV_COMMANDS[command]
            except KeyError:
                if errors:
                    path = _env_path.as_posix() if _env_path else "stream"
                    print(
                        f"unknown command {command} {path}({_lineno})",
                        file=sys.stderr,
                    )
    return _func, unquote(_key), unquote(_val)


def _process_stream(
    stream: BytesIO, environ, overwrite, errors, encoding=DEFAULT_ENCODING, env_path=None
):
    for lineno, line in enumerate(stream.readlines(), start=1):
        line = line.decode(encoding).strip()
        if line and line[0] != "#":
            func, key, val = _process_line(lineno, line, errors, env_path)
            if func is not None:
                func(environ, key, val, overwrite=overwrite)


def _process_env(
    env_file: str,
    search_path: List[Path],
    environ: MutableMapping[str, str],
    overwrite: bool,
    parents: bool,
    errors: bool,
    working_dirs: bool,
    decrypt: bool,
    password: Optional[str] = None,
    encoding: str = DEFAULT_ENCODING,
) -> MutableMapping[str, str]:
    """
    search for any env_files in the given dir list and populate environ dict

    :param env_file: base environment file name to use
    :param search_path: one or more paths to search
    :param environ: environment to update
    :param overwrite: whether to overwrite existing values
    :param parents: whether to search upwards until a file is found
    :param decrypt: whether to attempt decryption
    :param errors: whether to raise FileNotFoundError if the env_file is not found
    :param working_dirs: whether to add the env file's directory
    :param encoding: text encoding
    """
    files_not_found = []
    files_found = False
    for env_path in _env_files(env_file, search_path, parents, decrypt, errors):
        # insert PWD as the container of the env file
        env_path = Path(env_path).resolve()
        if working_dirs:
            environ["PWD"] = str(env_path.parent)
        try:
            with open_env(env_path) as f:
                data = f.read()
                if isinstance(data, str):
                    data = data.encode(encoding)
                load_stream(
                    BytesIO(data),
                    environ,
                    overwrite,
                    errors,
                    decrypt,
                    password,
                    encoding,
                    env_path,
                )
            files_found = True
        except FileNotFoundError:
            files_not_found.append(env_path)
    if errors and not files_found and files_not_found:
        raise FileNotFoundError(
            f"{env_file} as {[s.as_posix() for s in files_not_found]}"
        )
    return environ


def _process_var_reference(var_name: str, environ: MutableMapping[str, str]) -> str:
    """Process a variable reference and return its value or empty string if not found"""
    return environ.get(var_name, "")


def _process_shell_var(match_obj, environ: MutableMapping[str, str]) -> str:
    """
    Process shell-like variable substitution patterns:
    ${VAR} - Standard variable substitution
    ${VAR:-default} - Use default if VAR is not set
    ${VAR:+value} - Use value only if VAR is set
    """
    # Extract the full match and the variable name
    var_name = match_obj.group(1)

    # Check for modifiers
    if ":" in var_name:
        # Handle ${VAR:-default} or ${VAR:+value} patterns
        parts = _MODIFIER_PATTERN.split(var_name, maxsplit=1)
        if len(parts) == 3:
            var_name, modifier, value = parts

            # Process any nested variable references in the value
            value = _process_nested_vars(value, environ)

            var_value = _process_var_reference(var_name, environ)
            if modifier == "-" and not var_value or modifier == "+" and var_value:
                # Use default value if variable is not set
                return value
            elif modifier == "-":
                # Variable is set, use its value
                return var_value
            else:  # modifier == "+" and not var_value
                # Variable is not set, return empty string
                return ""

    # Standard variable substitution
    return _process_var_reference(var_name, environ)


MAX_RECURSION_DEPTH = 12


def _substitute_vars(value: str, environ: MutableMapping[str, str]) -> str:
    value = _VAR_BRACES_PATTERN.sub(lambda m: _process_shell_var(m, environ), value)
    value = _VAR_NO_BRACES_PATTERN.sub(
        lambda m: _process_var_reference(m.group(1), environ), value
    )
    return value


def _process_nested_vars(value: str, environ: MutableMapping[str, str]) -> str:
    for _ in range(MAX_RECURSION_DEPTH):
        new_value = _substitute_vars(value, environ)
        if new_value == value:
            break
        value = new_value
    return value


def _post_process(environ: MutableMapping[str, str]) -> MutableMapping[str, str]:
    """
    Post-process the variables using shell-like variable substitution:
    - ${VAR} - Standard variable substitution
    - ${VAR:-default} - Use default if VAR is not set
    - ${VAR:+value} - Use value only if VAR is set
    - $VAR - Variable substitution without braces
    """
    for key, val in environ.items():
        if "$" in val:  # Potential variable reference!
            new_val = _process_nested_vars(val, environ)
            if new_val != val:
                environ[key] = new_val
    return environ


def _update_os_env(environ: MutableMapping[str, str]) -> MutableMapping[str, str]:
    """back-populate changed variables to the environment"""
    for env_key, env_val in environ.items():
        if env_val != os.environ.get(env_key):
            os.environ[env_key] = env_val
    return os.environ


def load_env(
    env_file: str = None,
    search_path: Union[None, Union[List[str], List[Path]], str] = None,
    environ: MutableMapping[str, str] = None,
    overwrite: bool = False,
    parents: bool = False,
    update: bool = True,
    errors: bool = False,
    working_dirs: bool = True,
    decrypt: bool = False,
    password: str = None,
    encoding: Optional[str] = DEFAULT_ENCODING,
) -> MutableMapping[str, str]:
    """
    Loads one or more .env files with optional nesting, updating os.environ
    :param env_file: name of the environment file (.env or $ENV default)
    :param search_path: single or list of directories in order of precedence - str, bytes or Path
    :param environ: environment mapping to process
    :param overwrite: whether to overwrite existing values
    :param parents: whether to search upwards until a file is found
    :param update: update os.environ, default=True
    :param errors: whether to raise FileNotFoundError if env_file not found
    :param working_dirs: whether to add the env file's directory
    :param decrypt: whether to support encrypted .env.enc
    :param password: decryption password
    :param encoding: text encoding (default utf-8)
    :returns the new environment
    """
    if environ is None:
        environ = os.environ
    if not env_file:
        env_file = environ.get(DEFAULT_ENVKEY, DEFAULT_DOTENV)

    # insert this as a useful default
    if working_dirs:
        environ["CWD"] = Path.cwd().resolve(strict=True).as_posix()

    # determine where to search
    if search_path is None:
        import inspect

        frame = inspect.stack()[1]
        search_path = [".", frame.filename]
    elif isinstance(search_path, Path):
        search_path = [search_path]
    elif isinstance(search_path, (str, bytes)):
        search_path = search_path.split(os.pathsep)
    # convert to the array of Path for use internally
    search_path = [Path(p) for p in search_path]
    # if overwriting, traverse the path in reverse order so the first .env files have priority
    if overwrite:
        search_path.reverse()

    # slurp up the environment files found and
    # post-process values for template variables
    environ = _post_process(
        _process_env(
            env_file,
            search_path,
            environ.copy(),
            overwrite,
            parents,
            errors,
            working_dirs,
            decrypt,
            password,
            encoding,
        )
    )
    # optionally update the actual environment
    return _update_os_env(environ) if update else environ


def load_stream(
    stream: Union[BytesIO, TextIOBase],
    environ: MutableMapping[str, str] = None,
    overwrite: bool = False,
    errors: bool = False,
    decrypt: bool = False,
    password: Optional[str] = None,
    encoding: Optional[str] = DEFAULT_ENCODING,
    env_path: Optional[Path] = None,
):
    if isinstance(stream, TextIOBase):
        stream.seek(0)
        stream = BytesIO(stream.read().encode(encoding))
    elif password and decrypt:
        with contextlib.suppress(DecryptError):
            stream = decrypt_data(stream, password)
    _process_stream(stream, environ, overwrite, errors, encoding, env_path)


load_dotenv = load_env



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Populate vault with secrets from .env files and generate .env file with non-secrets.
Input:
  A .env file containing all values, secrets or otherwise.
  A "template" file that contains the variables to be exported,
   sensitive/secret values must be prefixed with a pipe | character
Output:

"""

import argparse
import re
import sys
from pathlib import Path
from string import Template
from typing import Union

from envex.dot_env import load_env
from envex.env_hvac import SecretsManager

SECRET_MARK = "|"


# noinspection DuplicatedCode
def read_env(
    envfile: Union[str, Path, None],
    search=None,
    parents=False,
    useenv=False,
    working_dirs=False,
):
    """Read the entire .env file"""
    if search is None:
        search_path = [Path.cwd()]
    else:
        search_path = set()
        for path in [p.split(",") for p in search]:
            for p in path:
                search_path.add(Path(p).resolve(strict=True))
        search_path = list(search_path)
    environ = None if useenv else {}
    return load_env(
        envfile,
        search_path=search_path,
        environ=environ,
        parents=parents,
        update=False,
        working_dirs=working_dirs,
    )


def cache_regex(rx: str) -> re.Pattern:
    if not hasattr(cache_regex, "cache"):
        cache_regex.cache = {}
    cache = cache_regex.cache
    if rx not in cache:
        cache[rx] = re.compile(rx)
    return cache[rx]


def env_match(var, regexlist, is_value=False):
    if regexlist:
        for regex in regexlist:
            if (
                is_value
                and var == regex
                or not is_value
                and cache_regex(regex).match(var)
            ):
                break
        else:
            return False
    return True


def subst(environ, lines) -> list:
    """post-process the variables using ${substitutions}"""
    data = []

    def do_subst(value: str) -> str:
        if all(v in value for v in ("${", "}")):  # looks like template
            # ignore anything that doesn't resolve, don't throw an exception.
            value = Template(value).safe_substitute(environ)
        return value

    for line in lines:
        if isinstance(line, tuple):
            var, val, secret = line[0], do_subst(line[1]), line[2]
            environ[var] = val  # update the environment
            data.append((var, val, secret))
        else:
            data.append(line)

    return data


def parse_template(env, template, with_comments=False):
    length = len(SECRET_MARK)
    lines = []
    with open(template, "r") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                if with_comments:
                    lines.append(line)
            else:
                secret = False
                if line.startswith(SECRET_MARK):
                    line = line[length:]
                    secret = True
                parts = line.split("=", maxsplit=1)
                if len(parts) == 1:
                    var = parts[0]
                    val = env.get(parts[0], "")
                else:
                    var, val = parts[0], parts[1]
                lines.append((var, val, secret))
    return lines


def output_result(lines: list, outputfile: str, empty: bool) -> dict:
    secrets = {}

    def writelines(fp, name, _lines):
        linecount = 0
        for line in _lines:
            if isinstance(line, tuple):
                var, val, secret = line[0], line[1], line[2]
                if val or empty:
                    if secret:
                        secrets[var] = val
                        continue
                    out = f"{var}={val}"
                else:
                    continue
            else:
                out = line
            print(out, file=fp)
            linecount += 1
        print(
            f"{name}: {linecount} line{'' if linecount == 1 else 's'} written",
            file=sys.stderr,
        )

    if outputfile == "-":
        writelines(sys.stdout, "<stdout>", lines)
    else:
        with open(outputfile, "w+") as f:
            writelines(f, outputfile, lines)

    return secrets


def secrets_manager(**kwargs) -> SecretsManager:
    sm = SecretsManager(**kwargs)
    if not sm.client:
        error("secrets manager not available")
    elif sm.sealed:
        error("secrets manager is sealed", 2)
    return sm


def create_or_update_secrets(secrets, key, cert, verbose):
    certpath = Path(cert).resolve().as_posix() if cert else None
    sm = secrets_manager(verify=certpath)
    sm.set_secrets(key, values=secrets)
    if verbose:
        import json

        data = sm.get_secrets(key) or secrets
        print(f"{sm.base_path}/{key}:")
        for s in json.dumps(data, indent=4).splitlines():
            print(f"    {s}")


def main(args):
    search = args.search.split(",") if args.search else None
    env = read_env(args.dotenv, search=search, parents=args.parents, useenv=args.environ)
    data = parse_template(env, args.template, args.comments)
    rendered = subst(env, data)
    if secrets := output_result(rendered, args.output, args.empty):
        create_or_update_secrets(secrets, args.key, args.cert, args.verbose)


def error(message, exitcode=None):
    print(f"{'ERROR' if exitcode else 'WARNING'}: {message}", file=sys.stderr)
    if exitcode is not None:
        exit(exitcode)


if __name__ == "__main__":
    prog = Path(sys.argv[0]).resolve(strict=True)
    parser = argparse.ArgumentParser(prog=prog.name, description=__doc__)

    scripts = Path(__file__).parent
    dotenv_default = ".env"
    template_default = scripts / "template.env"
    output_default = "docker.env"
    key_default = "app"

    parser.add_argument(
        "-e",
        "--environ",
        action="store_true",
        default=False,
        help="add OS environment to the list",
    )
    parser.add_argument(
        "-d",
        "--dotenv",
        action="store",
        default=dotenv_default,
        help=f"name of dot.env file (default={dotenv_default})",
    )
    parser.add_argument(
        "-s",
        "--search",
        action="store",
        nargs="?",
        help="search path(s) for env file (comma separated)",
    )
    parser.add_argument(
        "-p",
        "--parents",
        action="store_true",
        default=False,
        help="search parents until a dotenv file is found",
    )
    parser.add_argument(
        "-t",
        "--template",
        action="store",
        default=template_default,
        help=f'template file to use to use (default="{template_default}")',
    )
    parser.add_argument(
        "-c",
        "--comments",
        action="store_true",
        default=False,
        help="copy comments to output",
    )
    parser.add_argument(
        "-k",
        "--key",
        action="store",
        default=key_default,
        help=f'key for kv pairs stored in vault (default="{key_default}")',
    )
    parser.add_argument(
        "-E",
        "--empty",
        action="store_true",
        default=False,
        help="render or save empty values",
    )
    parser.add_argument(
        "-C",
        "--cert",
        action="store",
        default=None,
        help="path to Vault CA certificate chain file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="verbose output",
    )
    parser.add_argument(
        "output",
        action="store",
        default=output_default,
        help=f'output to this file (default="{output_default}")',
    )

    argv = parser.parse_args()
    # print(argv)

    main(argv)



import logging
from types import NoneType

__all__ = ("config", "log_set_level", "log_get_level")

from typing import NoReturn


def fatal(msg, *args, **kwargs) -> NoReturn:
    """
    Don't use this function, use critical() instead.
    """
    exitcode = kwargs.pop("exitcode", 1)
    logging.critical(msg, *args, **kwargs)
    exit(exitcode)


logging.fatal = fatal


def config(**kwargs):
    (kwargs.setdefault("level", logging.INFO),)
    (kwargs.setdefault("format", "%(asctime)s %(message)s (%(levelname)s)"),)
    logging.basicConfig(**kwargs)


__levelIndex = [
    logging.FATAL,
    logging.ERROR,
    logging.WARNING,
    logging.INFO,
    logging.DEBUG,
]


__default_level = logging.WARNING
__current_level = __default_level


def log_set_level(level: int | NoneType = None):
    """
    Set the log level for the application.

    :param level: The desired log level as an index.
                  If not specified, the level is reset to the default.
    :type level: int | NoneType
    :return: The loggging level that was set.
    :rtype: int
    """
    if level is None:
        level = __default_level
    else:
        if level < 0:
            level = 0
        elif level >= len(__levelIndex):
            level = len(__levelIndex) - 1
        level = __levelIndex[level]
    logging.getLogger().setLevel(level)
    return level


def log_get_level(level: int | NoneType = None):
    if level is None:
        level = __current_level
    return __levelIndex.index(level)



import argparse
from types import NoneType

__all__ = ("Decrement",)


# noinspection PyShadowingBuiltins
class Decrement(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest: str | NoneType,
        default: int = None,
        required: bool = False,
        help: str = None,
    ):
        super().__init__(
            option_strings, dest, nargs=0, default=default, required=required, help=help
        )

    # noinspection PyShadowingNames
    def __call__(self, parser, namespace, values, option_string=None):
        current_value = getattr(namespace, self.dest, self.default or 0)
        try:
            setattr(namespace, self.dest, current_value - 1)
        except TypeError:
            pass
